/*
 * HARK/hark-base
 *
 * Copyright (C) 2008-2018
 *
 */

#if !defined(HARK_BASE_MATRIX_H_INCLUDED)
#define HARK_BASE_MATRIX_H_INCLUDED 1

#include "common.h"

#include <complex>
#include "Object.h"
#include "ObjectParser.h"
#include "net_types.h"

namespace HARK
{
    class BaseMatrix : public Object
    {
    public:
	BaseMatrix() {}

	virtual size_t msize() const = 0;
	virtual bool mempty() const = 0;
	virtual ObjectRef clone() = 0;

    };

    template<class T>
    class Matrix : public Object //BaseMatrix
    {
    protected:
	int rows;
	int cols;
	T *data;

    public:

	using basicType = T;

	Matrix() :
	    rows(0),
	    cols(0),
	    data(NULL)
	{}

	Matrix(const Matrix &mat, bool transpose=false) :
	    rows(mat.rows),
	    cols(mat.cols),
	    data(new T [rows * cols])
	{
	    if (transpose) {
		rows=mat.cols;
		cols=mat.rows;
		for (int i=0; i<rows; i++)
		    for (int j=0; j<cols; j++) {
			data[i*cols+j] = mat.data[j*mat.cols+i];
		    }
	    } else {
		for (int i=0; i<rows*cols; i++) {
		    data[i] = mat.data[i];
		}
	    }
	}

	Matrix(int _rows, int _cols) :
	    rows(_rows),
	    cols(_cols),
	    data(new T [_rows*_cols])
	{}

	virtual ~Matrix()
	{
	    delete [] data;
	}

	void resize(int _rows, int _cols)
	{
	    T *new_data = new T [_rows*_cols];
	    int min_rows = _rows < rows ? _rows : rows;
	    int min_cols = _cols < cols ? _cols : cols;
	    for (int i=0; i<min_rows; i++)
		for (int j=0; j<min_cols; j++) {
		    new_data[i*_cols+j] = data[i*cols+j];
		}
	    if (data) {
		delete [] data;
	    }
	    data = new_data;
	    cols = _cols;
	    rows = _rows;
	}

	T *operator [] (int i)
	{
	    return data+i*cols;
	}

	const T *operator [] (int i) const
	{
	    return data+i*cols;
	}

	T &operator () (int i, int j)
	{
	    return data[i*cols+j];
	}

	const T &operator () (int i, int j) const
	{
	    return data[i*cols+j];
	}

	int nrows() const
	{
	    return rows;
	}

	int ncols() const
	{
	    return cols;
	}

	void transpose()
	{
	    if (rows==cols) {
		for (int i=0; i<rows; i++)
		    for (int j=0; j<i+1; j++) {
			float tmp=data[i*cols+j];
			data[i*cols+j] = data[j*cols+i];
			data[j*cols+i] = tmp;
		    }
	    } else {
		Matrix mat(*this);
		rows = mat.cols;
		cols = mat.rows;
		for (int i=0; i<rows; i++)
		    for (int j=0; j<cols; j++) {
			data[i*cols+j] = mat.data[j*mat.cols+i];
		    }
	    }
	}

	void printOn(std::ostream &out) const
	{
	    out << "<"<<className() << std::endl;
	    out << "<rows " << rows << ">" << std::endl;
	    out << "<cols " << cols << ">" << std::endl;
	    out << "<data " << std::endl;
	    for (int i=0; i<rows; i++) {
		for (int j=0; j<cols; j++) {
		    out << data[i*cols + j] << " ";
		}
		out << std::endl;
	    }
	    out << ">" << std::endl;
	    out << ">\n";
	}

	void readFrom(std::istream &in=std::cin);

	int size() const
	{
	    return (cols * rows);
	}

	virtual size_t msize() const
	{
	    return cols * rows;
	}

	virtual bool mempty() const
	{
	    return (cols == 0 && rows == 0);
	}

	static std::string GetClassName()
	{
	    std::string name = ObjectGetClassName<Matrix<T> >();
	    if (name == "unknown") {
		return std::string("Matrix");
	    } else {
		return name;
	    }
	}

	virtual ObjectRef clone();

    };

    template <class T>
    inline ObjectRef Matrix<T>::clone()
    {

	Matrix<T> *cpy = new Matrix<T>(this->nrows(), this->ncols());

	for (int i = 0; i < this->nrows(); i++) {
	    for (int j = 0; j < this->ncols(); j++) {
		(*cpy)(i,j) = (*this)(i,j);
	    }
	}
	return ObjectRef(cpy);
    }


    template <>
    inline ObjectRef Matrix<ObjectRef>::clone()
    {

	Matrix<ObjectRef> *cpy = new Matrix<ObjectRef>(this->nrows(), this->ncols());

	for (int i = 0; i < this->nrows(); i++) {
	    for (int j = 0; j < this->ncols(); j++) {
		//cloning every Object in the matrix
		(*cpy)(i,j) = (*this)(i,j)->clone();
	    }
	}

	return ObjectRef(cpy);
    }

    template <class T>
    inline void Matrix<T>::readFrom(std::istream &in)
    {
	std::string tag;
	int new_cols, new_rows;
	while (1) {
	    char ch;
	    in >> ch;
	    if (ch == '>') {
		break;
	    } else if (ch != '<') {
		throw new ParsingException ("Matrix<T>::readFrom : Parse error: '<' expected");
	    }
	    in >> tag;
	    if (tag == "rows") {
		in >> new_rows;
	    } else if (tag == "cols") {
		in >> new_cols;
	    } else if (tag == "data") {
		resize(new_rows,new_cols);
		for (int i=0; i<rows*cols; i++) {
		    in >> data[i];
		}
	    } else {
		throw new ParsingException ("Matrix<T>::readFrom : unknown argument: " + tag);
	    }

	    if (!in) {
		throw new ParsingException ("Matrix<T>::readFrom : Parse error trying to build " + tag);
	    }

	    in >> tag;
	    if (tag != ">") {
		throw new ParsingException ("Matrix<T>::readFrom : Parse error: '>' expected ");
	    }
	}

    }

}

#endif

/* end of file */
