/*
 * HARK/hark-base
 *
 * Copyright (C) 2008-2018
 *
 */

#if !defined(HARK_BASE_POINTER_H_INCLUDED)
#define HARK_BASE_POINTER_H_INCLUDED 1

#include "common.h"

#include "BaseException.h"
#include <typeinfo>

namespace HARK
{
    template <class T, class U>
    class PtrCastException : public GenericCastException
    {
    protected:
	std::string type;

    public:
	PtrCastException(const T *obj)
	    : type(typeid(*obj).name())
	{}

	virtual void print(std::ostream &out = std::cerr) const;
    };

    class PtrException : public BaseException
    {
    protected:
	std::string message;

    public:
	PtrException(const std::string &_message)
	    : message(_message)
	{}

	virtual void print(std::ostream &out = std::cerr) const;
    };

    template <class X>
    class RCPtr
    {
    public:
	using element_type = X;
	using pointer_type = X*;
	using size_type    = size_t;

    protected:
	X* ptr;

    public:
	explicit RCPtr(X* p=0) : ptr(p)
	{
	}

	bool isNil()
	{
	    return ptr == 0;
	}

	template <class Z>
	RCPtr(const RCPtr<Z> &r);

	RCPtr(const RCPtr<X> &r)
	{
	    ptr=r.ptr;
	    acquire();
	}

	~RCPtr()
	{
	    release();
	}

	template <class Z>
	RCPtr& operator= (const RCPtr<Z> &r);

	RCPtr& operator= (const RCPtr<X> &r)
	{
	    if (this != &r) {
		release();
		ptr = r.ptr;
		acquire();
	    }
	    return *this;
	}

	template <class Z>
	RCPtr& operator= (Z *r)
	{
	    if ((void*) ptr != (void*) (r)) {
		X *tmp=dynamic_cast<X*> (r);
		if (!tmp) {
		    throw new PtrCastException<Z,X>(r);
		}
		release();
		ptr=tmp;
	    }
	    return *this;
	}

	RCPtr& operator= (X *r)
	{
	    if (ptr != r) {
		release();
		ptr = r;
	    }
	    return *this;
	}

	X& operator* () const
	{
	    return *ptr;
	}

	X* operator->() const
	{
	    return  ptr;
	}

	X* get () const
	{
	    return ptr;
	}

	bool unique () const
	{
	    return (ptr && ptr->unique());
	}

	X *detach ()
	{
	    if (ptr) {
		if (ptr->unique()) {
		    X *tmp = ptr;
		    ptr = 0;
		    return tmp;
		} else {
		    throw new PtrException("Error: trying to detach a non-unique pointer in rc_ptrs.h");
		}
	    } else {
		throw new PtrException("Error: trying to detach a NULL pointer in rc_ptrs.h");
	    }
	}

    protected:
	void release()
	{
	    if (ptr) {
		ptr->unref();
	    }
	    ptr = 0;
	}

	void acquire()
	{
	    if (ptr) {
		ptr->ref();
	    }
	}

	template <class Z>
	friend class RCPtr;
    };

    template <class T, class U>
    void
    PtrCastException<T, U>::print(std::ostream &out) const
    {
	out << "Cast error: Trying to cast RCPtr <" << typeid(T).name() << "> (" << type << ") into RCPtr<" << typeid(U).name() << ">" << std::endl;
    }
    
    template<typename T>
    inline std::ostream &operator << (std::ostream &os, const RCPtr<T> &ptr)
    {
	return os << *ptr;
    }
}

#endif

/* end of file */
