/*
 * HARK/hark-base
 *
 * Copyright (C) 2008-2018
 *
 */

#include "Node.h"

#include "Object.h"

#include <typeinfo>
#include <sstream>
#include <string>

namespace HARK
{
    std::map<std::string,_NodeFactory*> &
    Node::factoryDictionary()
    {
	static std::map<std::string, _NodeFactory*> var;
	return var;
    }

    Node::Node()
    {
	throw new GeneralException("Node Constructor should not be called",__FILE__,__LINE__);
    }

    Node::Node(const std::string &nodeName, const ParameterSet &params) :
	name(nodeName),
	inputs(std::vector<NodeInput>()),
	parameters(params)
    {
    }

    Node::~Node()
    {}
    
    void
    Node::connectToNode(unsigned int in, Node *inputNode, unsigned int out)
    {
	if (inputs.size() <= in) {
	    char message[256];
	    throw new NodeException(this,message, __FILE__, __LINE__);
	} else {
	    inputs[in].outputID = out;
	    inputs[in].node = inputNode;
	}
    }

    int
    Node::addInput(const std::string &input_name)
    {
	std::vector<NodeInput>::iterator iter;

	for (iter = inputs.begin(); iter < inputs.end(); iter++) {
	    if (iter->name == input_name) {
		throw new NodeException(this,std::string("Input already defined : ") + input_name, __FILE__, __LINE__);
		break;
	    }
	}

	int position = inputs.size();
	inputs.resize(inputs.size() + 1);

	inputs[position] = NodeInput(input_name);

	return position;

    }

    int
    Node::addOutput(const std::string &output_name)
    {
	for (unsigned int in = 0; in < outputNames.size(); in++) {
	    if (outputNames[in] == output_name) {
		throw new NodeException(this,std::string("Output already defined : ") + output_name, __FILE__, __LINE__);
		break;
	    }
	}
	int position = outputNames.size();
	outputNames.resize(outputNames.size() + 1);
	outputNames[position] = output_name;
	return position;
    }

    bool
    Node::hasOutput(int output_id) const
    {
	return (output_id < int(outputNames.size()));
    }

    int
    Node::translateOutput(const std::string &output_name)
    {
	for (unsigned int in = 0; in < outputNames.size(); in++) {
	    if (outputNames[in] == output_name) {
		return in;
	    }
	}

	throw new NodeException(this,std::string("Unknown output in translateOutput : ") + output_name, __FILE__,__LINE__);

	return -1;
    }

    int
    Node::translateInput(const std::string &input_name)
    {
	for (unsigned int in = 0; in < inputs.size(); in++) {
	    if (inputs[in].name == input_name) {
		return in;
	    }
	}
	throw new NodeException(this,std::string("Unknown input in translateInput : ") + input_name, __FILE__,__LINE__);
	return -1;
    }

    int
    Node::getInputID(const std::string &inputName)
    {
	unsigned int i;
	for(i = 0; i < this->inputs.size(); i++){
	    if(this->inputs[i].name == inputName){
		return i;
	    }
	}
	return -1;
    }
    
    int
    Node::getOutputID(const std::string &outputName)
    {
	int id = -1;
	unsigned int i;
	for(i = 0; i < this->outputNames.size(); i++){
	    if(this->outputNames[i] == outputName){
		id = i;
		break;
	    }
	}
	//std::cout << "HARK::Node::getOutputID(" << outputName << ") -> " << id << std::endl;
	return id;
    }
    
    void
    Node::connectToNode(const std::string &in, Node *inputNode, const std::string &out)
    {
	connectToNode(translateInput(in), inputNode, inputNode->translateOutput(out));
    }

    void
    Node::initialize()
    {
    }

    _NodeFactory *
    Node::getFactoryNamed (const std::string &name)
    {

	_NodeFactory* factory = NULL;
	std::map<std::string, _NodeFactory*>::iterator iter;

	for (iter = factoryDictionary().begin(); iter != factoryDictionary().end(); iter++) {
	    if ((*iter).first == name) {
		factory = (*iter).second;
		break;
	    }
	}
	return factory;
    }

    int
    Node::addFactory(const std::string &factoryName, _NodeFactory* const factory)
    {
	if (!getFactoryNamed(factoryName)) {
	    if (factory != NULL) {
		factoryDictionary().insert (factoryEntry(factoryName,factory));
	    } else {
		std::cerr<<"Node::addFactory : NULL _NodeFactory pointer, exiting"<<std::endl;
		exit(-1);
	    }
	} else {
	    std::cerr << "Node::addFactory : The factory (" << factoryName << ") already exists. Ignoring the new one." << std::endl;
	}
	return 0;
    };

    void
    Node::printOn(std::ostream &out) const
    {
	out << "<Node" << std::endl;
	out << "<name " << name << " >" << std::endl;
	out << ">" << std::endl;
    }

    void
    NodeException::print(std::ostream &out) const
    {
	if (frozen) {
	    out << message;
	} else {
	    if (node) {
		out << file << " line " << line << ": Node " << node->getName()
		    << " (type " << typeid(*node).name() << ") " << message << std::endl;
	    } else {
		out << file << ", line " << line << ": " << message << std::endl;
	    }
	}
    }

    void
    NodeException::freeze()
    {
	if (frozen) {
	    return;
	}

	std::ostringstream outStr;
	print(outStr);
	message = outStr.str();
	frozen = true;
    }

    std::vector<NodeInput> &
    Node::getInputs()
    {
	return inputs;
    }

    ObjectRef
    Node::getOutputNamed(const std::string &outputName, int count)
    {
	return this->getOutput (this->translateOutput(outputName),count);
    }
    
    ObjectRef
    Node::getInput(int inputID, int count)
    {
	NodeInput &input = inputs[inputID];
	return input.node->getOutput(input.outputID, count);
    }

    std::string
    Node::getName() const
    {
	return name;
    }

    void
    Node::request(int outputID, const ParameterSet &req)
    {
	for (unsigned int i=0; i<inputs.size(); i++) {
	    inputs[i].node->request(inputs[i].outputID,req);
	}
    }
}

/* end of file */
